#!/usr/bin/env perl

#
# Copyright [2021-2022] International Business Machines Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

use strict;
use warnings;
use test;

test::ecsignverify("prime192v1", 8, 200, "SHA-256");
test::ecsignverify("secp224r1", 8, 200, "SHA-256");
test::ecsignverify("prime256v1", 8, 200, "SHA-256");
test::ecsignverify("secp384r1", 8, 200, "SHA-256");
test::ecsignverify("secp521r1", 8, 200, "SHA-256");
test::ecsignverify("brainpoolP160r1", 8, 200, "SHA-256");
test::ecsignverify("brainpoolP192r1", 8, 200, "SHA-256");
test::ecsignverify("brainpoolP224r1", 8, 200, "SHA-256");
test::ecsignverify("brainpoolP256r1", 8, 200, "SHA-256");
test::ecsignverify("brainpoolP320r1", 8, 200, "SHA-256");
test::ecsignverify("brainpoolP384r1", 8, 200, "SHA-256");
test::ecsignverify("brainpoolP512r1", 8, 200, "SHA-256");

